@@ -2,14 +2,15 @@
 using System;
 using System.Collections;
 using System.IO;
+using System.Runtime.InteropServices;
 using System.Xml;
 
 namespace Tomboy
 {
 	public delegate bool TagActivatedHandler (NoteTag tag,
-						  NoteEditor editor,
-						  Gtk.TextIter start, 
-						  Gtk.TextIter end);
+	                NoteEditor editor,
+	                Gtk.TextIter start,
+	                Gtk.TextIter end);
 
 	public class NoteTag : Gtk.TextTag
 	{
@@ -26,30 +27,30 @@ namespace Tomboy
 			CanActivate = 16,
 			CanSplit = 32
 		};
-		
+
 		TagFlags flags;
-		
+
 		public NoteTag (string tag_name)
-			: base(tag_name)
+: base(tag_name)
 		{
 			if (tag_name == null || tag_name == "") {
 				throw new Exception ("NoteTags must have a tag name.  Use " +
-						     "DynamicNoteTag for constructing " +
-						     "anonymous tags.");
+				                     "DynamicNoteTag for constructing " +
+				                     "anonymous tags.");
 			}
 
 			Initialize (tag_name);
 		}
 
 		internal NoteTag ()
-			: base (null)
+: base (null)
 		{
 			// Constructor used (only) by DynamicNoteTag
 			// Initialize() is called by NoteTagTable.Create().
 		}
 
 		public NoteTag (IntPtr raw)
-			: base (raw)
+: base (raw)
 		{
 			Logger.Log ("{0} IntPtr initializer called!", GetType());
 			Logger.Log ((new System.Diagnostics.StackTrace()).ToString());
@@ -64,67 +65,81 @@ namespace Tomboy
 
 		public string ElementName
 		{
-			get { return element_name; }
+			get {
+				return element_name;
+			}
 		}
 
-		public bool CanSerialize 
+		public bool CanSerialize
 		{
-			get { return (flags & TagFlags.CanSerialize) != 0; }
+			get {
+				return (flags & TagFlags.CanSerialize) != 0;
+			}
 			set {
 				if (value)
 					flags |= TagFlags.CanSerialize;
-				else 
+				else
 					flags &= ~TagFlags.CanSerialize;
 			}
 		}
 
-		public bool CanUndo 
+		public bool CanUndo
 		{
-			get { return (flags & TagFlags.CanUndo) != 0; }
+			get {
+				return (flags & TagFlags.CanUndo) != 0;
+			}
 			set {
 				if (value)
 					flags |= TagFlags.CanUndo;
-				else 
+				else
 					flags &= ~TagFlags.CanUndo;
 			}
 		}
 
 		public bool CanGrow
 		{
-			get { return (flags & TagFlags.CanGrow) != 0; }
+			get {
+				return (flags & TagFlags.CanGrow) != 0;
+			}
 			set {
 				if (value)
 					flags |= TagFlags.CanGrow;
-				else 
+				else
 					flags &= ~TagFlags.CanGrow;
 			}
 		}
 
 		public bool CanSpellCheck
 		{
-			get { return (flags & TagFlags.CanSpellCheck) != 0; }
+			get {
+				return (flags & TagFlags.CanSpellCheck) != 0;
+			}
 			set {
 				if (value)
 					flags |= TagFlags.CanSpellCheck;
-				else 
+				else
 					flags &= ~TagFlags.CanSpellCheck;
 			}
 		}
 
 		public bool CanActivate
 		{
-			get { return (flags & TagFlags.CanActivate) != 0; }
+			get {
+				return (flags & TagFlags.CanActivate) != 0;
+			}
 			set {
 				if (value)
 					flags |= TagFlags.CanActivate;
-				else 
+				else
 					flags &= ~TagFlags.CanActivate;
 			}
 		}
 
 		public bool CanSplit
 		{
-			get { return (flags & TagFlags.CanSplit) != 0; }
+			get {
+				return (flags & TagFlags.CanSplit) != 0;
+			}
 			set {
 				if (value)
 					flags |= TagFlags.CanSplit;
@@ -133,9 +148,9 @@ namespace Tomboy
 			}
 		}
 
-		public void GetExtents (Gtk.TextIter iter, 
-					out Gtk.TextIter start, 
-					out Gtk.TextIter end)
+		public void GetExtents (Gtk.TextIter iter,
+		                        out Gtk.TextIter start,
+		                        out Gtk.TextIter end)
 		{
 			start = iter;
 			if (!start.BeginsTag (this))
@@ -144,7 +159,7 @@ namespace Tomboy
 			end = iter;
 			end.ForwardToTagToggle (this);
 		}
-		
+
 		public virtual void Write (XmlTextWriter xml, bool start)
 		{
 			if (CanSerialize) {
@@ -165,9 +180,9 @@ namespace Tomboy
 			}
 		}
 
-		protected override bool OnTextEvent (GLib.Object  sender, 
-						     Gdk.Event    ev, 
-						     Gtk.TextIter iter)
+		protected override bool OnTextEvent (GLib.Object  sender,
+		                                     Gdk.Event    ev,
+		                                     Gtk.TextIter iter)
 		{
 			NoteEditor editor = (NoteEditor) sender;
 			Gtk.TextIter start, end;
@@ -184,7 +199,7 @@ namespace Tomboy
 
 				/* Don't activate if Shift or Control is pressed */
 				if ((int) (button_ev.State & (Gdk.ModifierType.ShiftMask |
-							      Gdk.ModifierType.ControlMask)) != 0)
+				                              Gdk.ModifierType.ControlMask)) != 0)
 					return false;
 
 				GetExtents (iter, out start, out end);
@@ -205,7 +220,7 @@ namespace Tomboy
 					return false;
 
 				if (key_ev.Key != Gdk.Key.Return &&
-				    key_ev.Key != Gdk.Key.KP_Enter)
+				                key_ev.Key != Gdk.Key.KP_Enter)
 					return false;
 
 				GetExtents (iter, out start, out end);
@@ -215,9 +230,9 @@ namespace Tomboy
 			return false;
 		}
 
-		protected virtual bool OnActivate (NoteEditor editor, 
-						   Gtk.TextIter start, 
-						   Gtk.TextIter end)
+		protected virtual bool OnActivate (NoteEditor editor,
+		                                   Gtk.TextIter start,
+		                                   Gtk.TextIter end)
 		{
 			bool retval = false;
 
@@ -242,7 +257,7 @@ namespace Tomboy
 				return image.Pixbuf;
 			}
 			set {
-				if(value == null) {
+				if (value == null) {
 					Widget = null;
 					return;
 				}
@@ -254,7 +269,9 @@ namespace Tomboy
 
 		public virtual Gtk.Widget Widget
 		{
-			get { return widget; }
+			get {
+				return widget;
+			}
 			set {
 				if (value == null && widget != null) {
 					widget.Destroy ();
@@ -279,8 +296,79 @@ namespace Tomboy
 
 		public virtual Gtk.TextMark WidgetLocation
 		{
-			get { return widgetLocation; }
-			set { widgetLocation = value; }
+			get {
+				return widgetLocation;
+			}
+			set {
+				widgetLocation = value;
+			}
+		}
+
+		// From contrast.h
+		public enum PaletteColor {
+			CONTRAST_COLOR_AQUA        =  0,
+			CONTRAST_COLOR_BLACK       =  1,
+			CONTRAST_COLOR_BLUE        =  2,
+			CONTRAST_COLOR_BROWN       =  3,
+			CONTRAST_COLOR_CYAN        =  4,
+			CONTRAST_COLOR_DARK_BLUE   =  5,
+			CONTRAST_COLOR_DARK_GREEN  =  6,
+			CONTRAST_COLOR_DARK_GREY   =  7,
+			CONTRAST_COLOR_DARK_RED    =  8,
+			CONTRAST_COLOR_GREEN       =  9,
+			CONTRAST_COLOR_GREY        = 10,
+			CONTRAST_COLOR_LIGHT_BLUE  = 11,
+			CONTRAST_COLOR_LIGHT_BROWN = 12,
+			CONTRAST_COLOR_LIGHT_GREEN = 13,
+			CONTRAST_COLOR_LIGHT_GREY  = 14,
+			CONTRAST_COLOR_LIGHT_RED   = 15,
+			CONTRAST_COLOR_MAGENTA     = 16,
+			CONTRAST_COLOR_ORANGE      = 17,
+			CONTRAST_COLOR_PURPLE      = 18,
+			CONTRAST_COLOR_RED         = 19,
+			CONTRAST_COLOR_VIOLET      = 20,
+			CONTRAST_COLOR_WHITE       = 21,
+			CONTRAST_COLOR_YELLOW      = 22,
+			CONTRAST_COLOR_LAST        = 23,
+		};
+
+		[DllImport("libtomboy")]
+		static extern Gdk.Color contrast_render_foreground_color(
+			        Gdk.Color background,
+			        PaletteColor symbol);
+
+		Gdk.Color get_background()
+		{
+			/* We can't know the exact background because we're not
+			   in TextView's rendering, but we can make a guess */
+			if (BackgroundSet)
+				return BackgroundGdk;
+
+			Gtk.Style s = Gtk.Rc.GetStyleByPaths(Gtk.Settings.Default,
+			                                     "GtkTextView", "GtkTextView", Gtk.TextView.GType);
+			if (s == null) {
+				Logger.Debug ("get_background: Style for GtkTextView came back null! Returning white...");
+				return new Gdk.Color (0xff, 0xff, 0xff); //white, for lack of a better idea
+			}
+			else
+				return s.Background(Gtk.StateType.Normal);
+		}
+
+		Gdk.Color render_foreground(PaletteColor symbol)
+		{
+			return contrast_render_foreground_color(get_background(), symbol);
+		}
+
+		private PaletteColor PaletteForeground_;
+		public PaletteColor PaletteForeground {
+			set {
+				PaletteForeground_ = value;
+				// XXX We should also watch theme changes.
+				ForegroundGdk = render_foreground(value);
+			}
+			get {
+				return PaletteForeground_;
+			}
 		}
 
 		public event Gtk.TagChangedHandler Changed;
@@ -291,16 +379,16 @@ namespace Tomboy
 		Hashtable attributes;
 
 		public DynamicNoteTag ()
-			: base()
+: base()
 		{
 		}
 
-		public Hashtable Attributes 
+		public Hashtable Attributes
 		{
-			get { 
+			get {
 				if (attributes == null)
 					attributes = new Hashtable ();
-				return attributes; 
+				return attributes;
 			}
 		}
 
@@ -330,34 +418,49 @@ namespace Tomboy
 						xml.ReadAttributeValue();
 						Attributes [name] = xml.Value;
 
+						OnAttributeRead (name);
 						Logger.Log (
-							"NoteTag: {0} read attribute {1}='{2}'",
-							ElementName,
-							name,
-							xml.Value);
+						        "NoteTag: {0} read attribute {1}='{2}'",
+						        ElementName,
+						        name,
+						        xml.Value);
 					}
 				}
 			}
 		}
+
+		/// <summary>
+		/// Derived classes should override this if they desire
+		/// to be notified when a tag attribute is read in.
+		/// </summary>
+		/// <param name="attributeName">
+		/// A <see cref="System.String"/> that is the name of the
+		/// newly read attribute.
+		/// </param>
+		protected virtual void OnAttributeRead (string attributeName) {}
 	}
-	
+
 	public class DepthNoteTag : NoteTag
 	{
 		int depth = -1;
 		Pango.Direction direction = Pango.Direction.Ltr;
-		
+
 		public int Depth
 		{
-			get{ return depth; }
+			get{
+				return depth;
+			}
 		}
-		
+
 		public Pango.Direction Direction
 		{
-			get{ return direction; }
+			get{
+				return direction;
+			}
 		}
 
 		public DepthNoteTag (int depth, Pango.Direction direction)
-			: base("depth:" + depth + ":" + direction)
+: base("depth:" + depth + ":" + direction)
 		{
 			this.depth = depth;
 			this.direction = direction;
@@ -368,7 +471,7 @@ namespace Tomboy
 			if (CanSerialize) {
 				if (start) {
 					xml.WriteStartElement (null, "list-item", null);
-					
+
 					// Write the list items writing direction
 					xml.WriteStartAttribute (null, "dir", null);
 					if (Direction == Pango.Direction.Rtl)
@@ -381,7 +484,7 @@ namespace Tomboy
 				}
 			}
 		}
-	}	
+	}
 
 	public class NoteTagTable : Gtk.TextTagTable
 	{
@@ -389,25 +492,25 @@ namespace Tomboy
 		Hashtable tag_types;
 		ArrayList added_tags;
 
-		public static NoteTagTable Instance 
+		public static NoteTagTable Instance
 		{
 			get {
-				if (instance == null) 
+				if (instance == null)
 					instance = new NoteTagTable ();
 				return instance;
 			}
 		}
 
-		public NoteTagTable () 
-			: base ()
+		public NoteTagTable ()
+: base ()
 		{
 			tag_types = new Hashtable ();
 			added_tags = new ArrayList ();
 
 			InitCommonTags ();
 		}
-		
-		void InitCommonTags () 
+
+		void InitCommonTags ()
 		{
 			NoteTag tag;
 
@@ -442,7 +545,7 @@ namespace Tomboy
 			Add (tag);
 
 			tag = new NoteTag ("highlight");
- 			tag.Background = "yellow";
+			tag.Background = "yellow";
 			tag.CanUndo = true;
 			tag.CanGrow = true;
 			tag.CanSpellCheck = true;
@@ -456,7 +559,8 @@ namespace Tomboy
 
 			tag = new NoteTag ("note-title");
 			tag.Underline = Pango.Underline.Single;
-			tag.Foreground = "#204a87";
+			tag.PaletteForeground =
+			        NoteTag.PaletteColor.CONTRAST_COLOR_BLUE;
 			tag.Scale = Pango.Scale.XXLarge;
 			// FiXME: Hack around extra rewrite on open
 			tag.CanSerialize = false;
@@ -472,7 +576,8 @@ namespace Tomboy
 			tag = new NoteTag ("datetime");
 			tag.Scale = Pango.Scale.Small;
 			tag.Style = Pango.Style.Italic;
-			tag.Foreground = "#888a85";
+			tag.PaletteForeground =
+			        NoteTag.PaletteColor.CONTRAST_COLOR_GREY;
 			Add (tag);
 
 			// Font sizes
@@ -509,19 +614,22 @@ namespace Tomboy
 
 			tag = new NoteTag ("link:broken");
 			tag.Underline = Pango.Underline.Single;
-			tag.Foreground = "#555753";
+			tag.PaletteForeground =
+			        NoteTag.PaletteColor.CONTRAST_COLOR_GREY;
 			tag.CanActivate = true;
 			Add (tag);
 
 			tag = new NoteTag ("link:internal");
 			tag.Underline = Pango.Underline.Single;
-			tag.Foreground = "#204a87";
+			tag.PaletteForeground =
+			        NoteTag.PaletteColor.CONTRAST_COLOR_BLUE;
 			tag.CanActivate = true;
 			Add (tag);
 
 			tag = new NoteTag ("link:url");
 			tag.Underline = Pango.Underline.Single;
-			tag.Foreground = "#3465a4";
+			tag.PaletteForeground =
+			        NoteTag.PaletteColor.CONTRAST_COLOR_BLUE;
 			tag.CanActivate = true;
 			Add (tag);
 		}
@@ -560,30 +668,30 @@ namespace Tomboy
 				return ((NoteTag) tag).CanActivate;
 			return false;
 		}
-		
+
 		public static bool TagHasDepth (Gtk.TextTag tag)
 		{
 			if (tag is DepthNoteTag)
 				return true;
-			
+
 			return false;
 		}
 
 		public DepthNoteTag GetDepthTag(int depth, Pango.Direction direction)
 		{
 			string name = "depth:" + depth + ":" + direction;
-			
+
 			DepthNoteTag tag = Lookup (name) as DepthNoteTag;
 
 			if (tag == null) {
 				tag = new DepthNoteTag (depth, direction);
 				tag.Indent = -14;
-				
+
 				if (direction == Pango.Direction.Rtl)
 					tag.RightMargin = (depth+1) * 25;
 				else
 					tag.LeftMargin = (depth+1) * 25;
-				
+
 				tag.PixelsBelowLines = 4;
 				tag.Scale = Pango.Scale.Medium;
 				Add (tag);
@@ -595,7 +703,7 @@ namespace Tomboy
 		public DynamicNoteTag CreateDynamicTag (string tag_name)
 		{
 			Type tag_type = tag_types [tag_name] as Type;
-			if (tag_type == null) 
+			if (tag_type == null)
 				return null;
 
 			DynamicNoteTag tag = (DynamicNoteTag) Activator.CreateInstance(tag_type);
